package org.processmining.streamheuristicsminer.algorithms;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;

import javax.naming.OperationNotSupportedException;

import org.deckfour.xes.classification.XEventClass;
import org.processmining.acceptingpetrinet.models.AcceptingPetriNet;
import org.processmining.acceptingpetrinet.models.impl.AcceptingPetriNetFactory;
import org.processmining.dataawarecnetminer.converter.CausalNetToPetrinet;
import org.processmining.datapetrinets.DataPetriNet.PetrinetWithMarkings;
import org.processmining.eventstream.core.interfaces.XSEvent;
import org.processmining.eventstream.readers.abstr.AbstractXSEventReader;
import org.processmining.eventstream.readers.acceptingpetrinet.AcceptingPetriNetXSReaderResultVisualizer;
import org.processmining.eventstream.readers.acceptingpetrinet.XSEventStreamToAcceptingPetriNetReader;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.util.Pair;
import org.processmining.logabstractions.models.DirectlyFollowsAbstraction;
import org.processmining.models.cnet.CNet;
import org.processmining.stream.algorithms.lossycounting.CNetGenerator;
import org.processmining.stream.model.datastructure.DSParameterMissingException;
import org.processmining.stream.model.datastructure.DSWrongParameterException;
import org.processmining.stream.model.datastructure.DataStructureFactory;
import org.processmining.stream.model.datastructure.PointerBasedDataStructure;
import org.processmining.stream.reader.views.XSReaderSequenceOfResultsView;
import org.processmining.streamabstractrepresentation.algorithms.UncoupledActivityFrequencyApproximationReaderImpl;
import org.processmining.streamabstractrepresentation.algorithms.UncoupledXEventClassDFAReaderImpl;
import org.processmining.streamheuristicsminer.parameters.UncoupledStreamHeuristicMinerParametersImpl;

import gnu.trove.map.TObjectLongMap;

public class UncoupledAcceptingPetriNetStreamHeursticsMinerImpl
		extends AbstractXSEventReader<AcceptingPetriNet, AcceptingPetriNet, UncoupledStreamHeuristicMinerParametersImpl>
		implements XSEventStreamToAcceptingPetriNetReader {

	private PointerBasedDataStructure<String, XEventClass> caseAdministration;

	private UncoupledActivityFrequencyApproximationReaderImpl<Object, XEventClass, UncoupledStreamHeuristicMinerParametersImpl> activityApproxDelegate;
	private UncoupledXEventClassDFAReaderImpl<Object, UncoupledStreamHeuristicMinerParametersImpl> dfaDelegate;

	public UncoupledAcceptingPetriNetStreamHeursticsMinerImpl(final PluginContext context,
			UncoupledStreamHeuristicMinerParametersImpl parameters) {
		super("stream_heuristics_apn_ca_" + parameters.getCaseAdministrationDataStructureType().getName() + "-"
				+ parameters.getCaseAdministrationDataStructureParameters().toString() + "aa_"
				+ parameters.getInternalRepresentationDataStructureType().getName() + "-"
				+ parameters.getInternalRepresentationDataStructureParameters().toString(),
				new XSReaderSequenceOfResultsView<>("Visualizer of: Stream Heuristics Miner (Accepting Petri Net)",
						null, parameters.getRefreshRate(), AcceptingPetriNetXSReaderResultVisualizer.instance(context)),
				parameters);

		((XSReaderSequenceOfResultsView<AcceptingPetriNet>) getVisualization()).setReader(this);

		try {
			caseAdministration = DataStructureFactory.createPointerDataStructure(
					parameters.getCaseAdministrationDataStructureType(),
					parameters.getCaseAdministrationDataStructureParameters());
		} catch (OperationNotSupportedException | DSParameterMissingException | DSWrongParameterException e) {
			e.printStackTrace();
			caseAdministration = null;
			interrupt();
		}
		dfaDelegate = new UncoupledXEventClassDFAReaderImpl<Object, UncoupledStreamHeuristicMinerParametersImpl>(
				"stream_heuristics_dfa_delegate", null, parameters, caseAdministration, false);

		activityApproxDelegate = new UncoupledActivityFrequencyApproximationReaderImpl<Object, XEventClass, UncoupledStreamHeuristicMinerParametersImpl>(
				"stream_heuristics_approx_acitvity_delegate", null, parameters, caseAdministration, false);
	}

	public void processNewXSEvent(String caseId, XEventClass activity) {
		caseAdministration.add(caseId, activity);

	}

	protected AcceptingPetriNet computeCurrentResult() {
		HashMap<String, Double> activityCounts = convertActivityApprox();
		DirectlyFollowsAbstraction<XEventClass> dfa = dfaDelegate.getCurrentResult();
		HashMap<Pair<String, String>, Double> dfr = convertDFA(dfa);
		HashMap<String, Integer> startingActivities = deduceStartingActivities(dfa);
		HashMap<String, Integer> finishingActivities = deduceEndActivities(dfa);
		CNetGenerator generator = new CNetGenerator(activityCounts, dfr, startingActivities, finishingActivities);
		CNet cNet = generator.generateModel(getParameters().getDependencyThreshold(),
				getParameters().getPositiveObservations(), getParameters().getAndThreshold());
		CausalNetToPetrinet converter = new CausalNetToPetrinet();
		PetrinetWithMarkings pnwm = converter.convert(cNet);
		AcceptingPetriNet apn = AcceptingPetriNetFactory.createAcceptingPetriNet(pnwm.getNet(),
				pnwm.getInitialMarking(), pnwm.getFinalMarkings());
		return apn;
	}

	private HashMap<String, Integer> deduceStartingActivities(final DirectlyFollowsAbstraction<XEventClass> dfa) {
		HashMap<String, Integer> startActivities = new HashMap<>();
		Set<XEventClass> sources = new HashSet<>();
		// should never be a target
		for (XEventClass target : dfa.getEventClasses()) {
			boolean isTarget = false;
			for (XEventClass source : dfa.getEventClasses()) {
				if (dfa.getValue(source, target) > 0) {
					isTarget = true;
					break;
				}
			}
			if (!isTarget) {
				sources.add(target);
			}
		}

		for (XEventClass source : sources) {
			startActivities.put(source.toString(), 0);
			for (XEventClass target : dfa.getEventClasses()) {
				startActivities.put(source.toString(),
						startActivities.get(source.toString()) + new Double(dfa.getValue(source, target)).intValue());
			}
		}
		return startActivities;
	}

	private HashMap<String, Integer> deduceEndActivities(final DirectlyFollowsAbstraction<XEventClass> dfa) {
		HashMap<String, Integer> endActivities = new HashMap<>();
		Set<XEventClass> targets = new HashSet<>();
		// should never be a source
		for (XEventClass source : dfa.getEventClasses()) {
			boolean isSource = false;
			for (XEventClass target : dfa.getEventClasses()) {
				if (dfa.getValue(source, target) > 0) {
					isSource = true;
					break;
				}
			}
			if (!isSource) {
				targets.add(source);
			}
		}

		for (XEventClass target : targets) {
			endActivities.put(target.toString(), 0);
			for (XEventClass source : dfa.getEventClasses()) {
				endActivities.put(target.toString(),
						endActivities.get(target.toString()) + new Double(dfa.getValue(source, target)).intValue());
			}
		}
		return endActivities;
	}

	private HashMap<Pair<String, String>, Double> convertDFA(final DirectlyFollowsAbstraction<XEventClass> dfa) {
		HashMap<Pair<String, String>, Double> res = new HashMap<>();
		for (XEventClass e1 : dfa.getEventClasses()) {
			for (XEventClass e2 : dfa.getEventClasses()) {
				res.put(new Pair<String, String>(e1.toString(), e2.toString()), dfa.getValue(e1, e2));
			}
		}
		return res;
	}

	private HashMap<String, Double> convertActivityApprox() {
		HashMap<String, Double> res = new HashMap<>();
		TObjectLongMap<XEventClass> approx = activityApproxDelegate.getCurrentResult();
		for (XEventClass e : approx.keySet()) {
			res.put(e.toString(), (double) approx.get(e));
		}
		return res;
	}

	@Override
	public void handleNextPacket(XSEvent event) {
		if (canReadPacket(event)) {
			dfaDelegate.handleNextPacket(event);
			activityApproxDelegate.handleNextPacket(event);
			super.handleNextPacket(event);
		}

	}

}
